#!/usr/bin/env python
#-*-coding: utf-8 -*-

import os, sys, json
from struct import pack, unpack
from time import strptime, mktime
from syslog import syslog, LOG_ERR

PY3 = sys.version_info[0] == 3

if PY3:
    from io import StringIO
else:
    from StringIO import StringIO

debug = False
def log_debug(*args):
	if debug:
		print >> sys.stderr, ' '.join(args)
def _is_scoket_exception(e):
	import socket
	if type(e) == socket.gaierror or type(e) == socket.error or type(e) == socket.herror or type(e) == socket.timeout:
		return True
	return False

def _convert_socket_error_code(message):
	error_code = -1
	if -1 != message.find("[Errno -2] Name or service not known"):
		error_code = -2
	elif -1 != message.find("[Errno -3] Temporary failure in name resolution"):
		error_code = -2
	elif -1 != message.find("[Errno 110] Connection timed out"):
		error_code = 408
	elif -1 != message.find("[Errno 101] Network is unreachable"):
		error_code = -4
	elif -1 != message.find("[Errno 111] Connection refused"):
		error_code = -4
	elif -1 != message.find("[Errno 113] No route to host"):
		error_code = -4
	elif -1 != message.find("[Errno 104] Connection reset by peer"):
		error_code = -4
	elif -1 != message.find("[Errno 32] Broken pipe"):
		error_code = -4
	return error_code

def _convert_rest_socket_error_code(message):
        error_code = -1
        #ssl.SSLError: SSL exception log
        if -1 != message.find("The write operation timed out"):
                error_code = 408
        elif -1 != message.find("The read operation timed out"):
                error_code = 408
        else:
                error_code = _convert_socket_error_code(message)
        return error_code

def _convert_protocol_error_code(message):
        error_code = -1
        if -1 != message.find("BadStatusLine"):
                error_code = -4
        return error_code

def _convert_socket_exception(e):
	import socket

	error_code = -1
	error_msg = 'Unknown'
	error_cls = 'Unknown'

	try:
		error_cls = type(e).__name__
		if type(e) == socket.gaierror:
			if hasattr(e, 'strerror') and e.strerror:
				error_msg = e.strerror
			if e.errno == -2:
				# socket.gaierror: [Errno -2] Name or service not known
				# 1. dns not set
				# 2. set to a wrong dns
				error_code = -2
			elif e.errno == -3:
				# socket.gaierror: [Errno -3] Temporary failure in name resolution
				# man gai_strerror, and find EAI_AGAIN:
				# The name server returned a temporary failure indication.  Try again later.
				error_code = -2
			elif e.errno < 100 and e.errno > 0:
				error_code = e.errno
		elif type(e) == socket.error or type(e) == socket.herror:
			if hasattr(e, 'strerror') and e.strerror:
				error_msg = e.strerror
			if e.errno == socket.errno.ETIMEDOUT:
				# socket.error: [Errno 110] Connection timed out
				error_code = 408
			elif e.errno == socket.errno.ENETUNREACH:
				# socket.error: [Errno 101] Network is unreachable
				error_code = -4
			elif e.errno == socket.errno.ECONNREFUSED:
				# socket.error: [Errno 111] Connection refused
				error_code = -4
			elif e.errno == socket.errno.EHOSTUNREACH:
				# socket.error: [Errno 113] No route to host
				error_code = -4
			elif e.errno == socket.errno.ECONNRESET:
				# socket.error: [Errno 104] Connection reset by peer
				error_code = -4
			elif e.errno == socket.errno.EPIPE:
				# socket.error: [Errno 32] Broken pipe
				error_code = -4
			elif e.errno < 100 and e.errno > 0:
				error_code = e.errno
			#syslog(LOG_ERR, "get network error %d:%s" % (e.errno, error_msg))
		elif type(e) == socket.timeout:
			error_msg = 'timed out'
			error_code = 408
		#syslog(LOG_ERR, "error %s:%d:%s" % (error_cls, error_code, str(error_msg)))
	except:
		#syslog(LOG_ERR, "parse socket exception failed")
		pass

	return {
		'success': False,
		'error_class': error_cls,
		'error_message': error_msg,
		'error_code': error_code
	}

def _convert_exception(e):
        #syslog(LOG_ERR, "convert_exception: %s" % str(e))
	import dropbox
	import urllib3
	import socket
	#import pdb; pdb.set_trace()

	error_code = -1
	error_msg = 'Unknown'
	error_cls = 'Unknown'

	try:
		error_cls = type(e).__name__

		try:
			if type(e.message) is unicode:
				error_msg = repr(e.message).split('\\n')[0]
			else:
				error_msg = str(e.message).split('\n')[0]
			log_debug("\033[31mexception: \033[0m", error_msg)
		except:
			log_debug("\033[31mexception: \033[0m", "could not parse error msg")

                if type(e) == dropbox.rest.ErrorResponse:
                        if hasattr(e, 'status') and e.status:
    	        		error_code = e.status
                        if hasattr(e, 'error_msg') and e.error_msg:
                                error_msg = e.error_msg
                elif type(e) == dropbox.rest.RESTSocketError:
                                error_code = _convert_rest_socket_error_code(str(e))
		elif type(e) == TypeError:
			# this exception will raise when base64 decode for auth failed
			if error_msg == "Incorrect padding":
				error_code = 401
		elif type(e) == UnicodeDecodeError:
			error_code = -5
                elif type(e) == urllib3.exceptions.ProtocolError:
			if _is_scoket_exception(e.args[1]):
				return _convert_socket_exception(e.args[1])
                elif type(e) == urllib3.exceptions.MaxRetryError:
        		if type(e.reason) == urllib3.exceptions.ReadTimeoutError:
                                error_code = 408
                        elif type(e.reason) == urllib3.exceptions.TimeoutError:
                                error_code = 408
                        elif type(e.reason) == urllib3.exceptions.NewConnectionError:
                                error_code = _convert_socket_error_code(str(e.reason))
                        elif type(e.reason) == urllib3.exceptions.ProtocolError:
                                error_code = _convert_protocol_error_code(str(e.reason))
                elif type(e) == urllib3.exceptions.ConnectTimeoutError:
                        error_code = 408
		elif type(e) == urllib3.exceptions.NewConnectionError:
			error_code = _convert_socket_error_code(str(e))
		elif _is_scoket_exception(e):
			return _convert_socket_exception(e)
                else:
                    if not error_msg:
                        error_msg = str(e)

		#syslog(LOG_ERR, "error %s:%d:%s" % (error_cls, error_code, str(error_msg)))
	except Exception as e:
		#syslog(LOG_ERR, "parse exception failed. %s" % str(e))
                pass
	return {
		'success': False,
		'error_class': error_cls,
		'error_message': error_msg,
		'error_code': error_code
	}
def _convert_time_str(val):
        if val is None:
            return 0
        strVal = str(val)
	origin_tz = None
	if os.environ.get('TZ'):
		origin_tz = os.environ.get('TZ')
	os.environ['TZ'] = 'GMT'

        # Sat, 21 Aug 2010 22:31:20 +0000
        timestamp = int(mktime(strptime(strVal[:25], '%a, %d %b %Y %H:%M:%S')))

	if origin_tz:
		os.environ['TZ'] = origin_tz
	return timestamp
def _convert_properties(properties):
	out_json = {
		'LastModified': _convert_time_str(properties.get('modified')),
		'ContentLength': properties.get('bytes'),
                'is_dir': properties.get('is_dir'),
                'path': properties.get('path'),
                'is_deleted': properties.get('is_deleted', False),
                'rev': properties.get('rev')
	}
	return out_json

class SimpleIO(object):
	def read_int(self):
		data = ''
		n = 4
		while n > 0:
			read_data = sys.stdin.read(n)
			if 0 == len(read_data):
				raise StopIteration
			data += read_data
			n -= len(read_data)
		if data:
			return unpack('i', data)[0]
		else:
			# check eof?
			raise StopIteration
	def read_string(self):
		n = self.read_int()
		if 0 == n:
			return ''
		data = ''
		while n > 0:
			read_data = sys.stdin.read(n)
			if 0 == len(read_data):
				raise StopIteration
			data += read_data
			n -= len(read_data)
		if data:
			return data
		else:
			raise SystemError
	def read_json(self):
		json_str = self.read_string()
		if json_str:
			return json.loads(json_str)
		else:
			return None
	def write_int(self, val):
		data = pack('i', val)
		sys.stdout.write(data)
		sys.stdout.flush()
	def write_string(self, val):
		self.write_int(len(val))
		sys.stdout.write(val)
		sys.stdout.flush()
		# log
	def write_json(self, val):
		s = json.dumps(val)
		self.write_string(s)
	def write_exception(self, e):
		self.write_json(_convert_exception(e))

class FileObjectWithProgress(object):
	"""
	Readable file object with progress wrapper.
	"""

	def __init__(self, io):
		"""
		Wrap the underlying file object

		:param file_obj: the file object to wrap
		"""
		self._io = io
		self._file_obj = None
		self._prev_process_size = 0

        def __getattr__(self, attr):
                return getattr(self._file_obj, attr)

	def set_file_obj(self, file_obj):
		self._file_obj = file_obj
		self._prev_process_size = 0

	def read(self, length=None):
		read_data = self._file_obj.read(length)
		self._prev_process_size += len(read_data)

		res = {
			'success': True,
			'complete': False,
			'uploaded': self._prev_process_size
		}
		if self._io:
			self._io.write_json(res)
		else:
			print res
		return read_data

	def write(self, write_data):
		self._file_obj.write(write_data)
		self._prev_process_size += len(write_data)

		res = {
			'success': True,
			'complete': False,
			'downloaded': self._prev_process_size
		}
		if self._io:
			self._io.write_json(res)
		else:
			print res
		return None
	def tell(self):
		return self._file_obj.tell();
	def seek(self, offset, whence=os.SEEK_SET):
		self._file_obj.seek(offset, whence)

class DropboxStoreageApi(object):
	def __init__(self,io=None):
                from dropbox import client, rest, session

		self._file_obj_progress = FileObjectWithProgress(io)
                self.prompt = "Dropbox> "
                access_token = os.environ.get('DROPBOX_ACCESS_TOKEN')
                #syslog(LOG_ERR, "access_token: %s" % str(access_token))
                self.api_client = client.DropboxClient(access_token)
                del os.environ['DROPBOX_ACCESS_TOKEN']
        def metaData(self, in_json):
            #syslog(LOG_ERR, "metaData: %s" % str(in_json))
            path = in_json.get('path', None)
            listChild = in_json.get('listChild', True)
            metadata = self.api_client.metadata(path, listChild, file_limit=25000, hash=None, rev=None, include_deleted=False, include_media_info=False)
            # 304: Current folder hash matches hash parameters, so contents are unchanged.
            # 400: Bad request (may be due to many things; check e.error for details).
            # 404: No file was found at given path.
            # 406: Too many file entries to return.

            count = 0
            out_json = {
                    'success': True,
                    'Name': os.path.basename(metadata['path']),
                    'Properties': _convert_properties(metadata),
                    'child': []
            }

            if 'contents' in metadata:
                for f in metadata['contents']:
                    #print f['path']

                    out_json['child'].append({
                        'Name': os.path.basename(f['path']),
                        'Properties': _convert_properties(f)
                        })
                    count += 1

            out_json['count'] = count
            #print out_json
            return out_json

        def createDir(self, in_json):
            #print 'createDir:', in_json
            path = in_json.get('path', None)
            #print path
            metadata = self.api_client.file_create_folder(path)
            # 403   There is already a file or folder at the given destination.

            out_json = {
                    'success': True,
                    'Name': os.path.basename(metadata['path']),
                    'Properties': _convert_properties(metadata)
            }
            return out_json

        def removeObject(self, in_json):
            #print 'removeObject:', in_json
            path = in_json.get('path', None)
            #print path
            metadata = self.api_client.file_delete(path)
            # 404   No file was found at the specified path.
            # 406   Too many files would be involved in the operation for it to complete successfully.
            #       The limit is currently 10,000 files and folders.

            out_json = {
                    'success': True,
                    'Name': os.path.basename(metadata['path']),
                    'Properties': _convert_properties(metadata)
            }
            return out_json

        def getAccountInfo(self, in_json):
            f = self.api_client.account_info()
            out_json = {
                    'success': True,
                    'Properties': []
            }
            out_json['Properties'] =  f
            return out_json

        def putFile(self, in_json):
            from_path = in_json.get('from_path')
            to_path = in_json.get('to_path')

            with  open(from_path, "rb") as from_file:
		self._file_obj_progress.set_file_obj(from_file)
                metadata = self.api_client.put_file(to_path, self._file_obj_progress, overwrite=True)

            # 400: Bad request (may be due to many things; check e.error for details).
            # 507: User over storage quota.

            out_json = {
                    'success': True,
                    'Name': os.path.basename(metadata['path']),
                    'Properties': _convert_properties(metadata)
            }
            return out_json

        def getFileAndMeta(self, in_json):
            from_path = in_json.get('from_path')
            to_path = in_json.get('to_path')

            with open(to_path, "wb") as to_file:
		self._file_obj_progress.set_file_obj(to_file)
                f, metadata = self.api_client.get_file_and_metadata(from_path)
                self._file_obj_progress.write(f.read())
            # 400: Bad request (may be due to many things; check e.error for details).
            # 404: No file was found at the given path, or the file that was there was deleted.
            # 200: Request was okay but response was malformed in some way.

            #print 'Metadata:', metadata
            #print 'f:', f
            out_json = {
                    'success': True,
                    'Name': os.path.basename(metadata['path']),
                    'Properties': _convert_properties(metadata)
            }
            return out_json


        def putChunk(self, in_json):
            from_path = in_json.get('from_path')
            length = int(in_json.get('length'))
            offset = int(in_json.get('offset', 0))
            upload_id = in_json.get('upload_id', None)
            #print "off_set:",offset, " upload_id:", upload_id
            with open(from_path, "rb") as from_file:
		self._file_obj_progress.set_file_obj(from_file)
                self._file_obj_progress.seek(offset)
                new_offset, upload_id = self.api_client.upload_chunk(StringIO(self._file_obj_progress.read(length)),
                                                                     length, offset, upload_id)
            # 404   The upload_id does not exist or has expired.
            # 400   The offset parameter does not match up with what the server expects.
            #       The body of the error response will be JSON similar to the above, indicating the correct offset to upload.
            out_json = {
                    'success': True,
                    'upload_id': upload_id,
                    'new_offset': new_offset
            }
            return out_json

        def commitChunks(self, in_json):
            to_path = "auto/" + in_json.get('to_path')
            upload_id = in_json.get('upload_id')
            metadata = self.api_client.commit_chunked_upload(to_path, upload_id, overwrite=True)
            # 409   The call failed because a conflict occurred.
            #       This means a file already existed at the specified path,
            #       overwrite was false, and the parent_rev (if specified) didn't match the current rev.
            # 400   Returned if the request does not contain an upload_id
            #       or if there is no chunked upload matching the given upload_id.
            out_json = {
                    'success': True,
                    'Name': os.path.basename(metadata['path']),
                    'Properties': _convert_properties(metadata)
            }
            return out_json

        def getFileChunk(self, in_json):
            from_path = in_json.get('from_path')
            to_path = in_json.get('to_path')
            start = in_json.get('start', None)
            length = int(in_json.get('length', None))

            with open(to_path, "ab") as to_file:
		self._file_obj_progress.set_file_obj(to_file)
                if start is not None:
                    self._file_obj_progress.seek(start)
                f = self.api_client.get_file(from_path, None, start, length)
                self._file_obj_progress.write(f.read())
            # 400: Bad request (may be due to many things; check e.error for details).
            # 404: No file was found at the given path, or the file that was there was deleted.
            # 200: Request was okay but response was malformed in some way.

            #print 'Metadata:', metadata
            #print 'f:', f
            out_json = {
                    'success': True,
            }
            return out_json

        def delta(self, in_json):
            cursor = in_json.get('cursor', None)
            path = in_json.get('path', None)
            result = self.api_client.delta(cursor, path, include_media_info=False)
            #print result
            count = 0
            files = {}
            out_json = {
                    'success': True,
                    'cursor': result.get('cursor'),
                    'has_more': result.get('has_more'),
                    'entries': []
            }
            #if not exist metadata is none, return count = 0
            #if only file/folder itself, return count =1
            #others return all file and itself
            for lowercase_path, metadata in result.get('entries'):
                    if metadata is not None:
                        out_json['entries'].append({
                            'lowercase_path': lowercase_path,
                            'Properties': _convert_properties(metadata)
                        })
                        count += 1

            out_json['count'] = count
            #print out_json
            return out_json




def start_server():
	io = SimpleIO()

	try:
		api = DropboxStoreageApi(io)
	except Exception as e:
		io.write_exception(e)
		return False

	io.write_string('start')

	while True:
		try:
			in_json = io.read_json()
			fn_name = in_json['fn']
			fn = getattr(api, fn_name)
			del in_json['fn']

			if fn is None:
				raise SystemError('no such fn: ' + fn_name)

			#log_debug('\033[34mexecute: ' + fn_name + ' ' + json.dumps(in_json) + '\033[0m');

			res = fn(in_json)
			io.write_json(res)
		except StopIteration:
			break
		except Exception as e:
			io.write_exception(e)
			continue

if __name__ == '__main__':
	# add include path
	script_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
	sys.path.insert(0, script_dir + '/dropbox')
	sys.path.insert(1, script_dir + '/module')

	res = start_server()
	sys.exit(0 if res else 1)
